import type { Coordinate } from './Coordinate';
import { Row } from './Row';
import { Column } from './Column';
import type { Cell } from './Cell';
import { type CellValueChangedEvent } from './utils';
import type { StraightDirection } from './Direction';
/**
 * Options to generate a new grid
 */
export type InitializeNewGridOptions<Value> = {
    /**
     * The amount of columns (min: 1)
     */
    width: number;
    /**
     * The amount of rows (min: 1)
     */
    height: number;
    /**
     * The given function will be called, when the cell at {coordinate} will be initialized. The returned value will be the value of the cell.
     * @param coordinate The coordinate of the cell that is being initialized
     */
    initializeCellValue: (coordinate: Coordinate) => Value;
};
/**
 * Options to generate a new grid based on an existing 2-dimensional array.
 */
export type PreInitializedGridOptions<Value> = {
    /**
     * The 2-dimensional array representing a grid (1. dimension: row (y-axis), 2. dimension: column (x-axis))
     */
    grid: Value[][];
};
export type InitializeGridOptions<Value> = InitializeNewGridOptions<Value> | PreInitializedGridOptions<Value>;
/**
 * The grid contains all information about cells
 */
export declare abstract class Grid<Value, CellWithValue extends Cell<Value>> {
    private options;
    readonly width: number;
    readonly height: number;
    private readonly _grid;
    private readonly _rows;
    private readonly _columns;
    private readonly eventDispatcher;
    /**
     * @param options The initialization configuration
     */
    protected constructor(options: InitializeGridOptions<Value>);
    protected initialize(initializeCell: (coordinate: Coordinate, value: Value) => CellWithValue): void;
    private initializeCells;
    /**
     * @returns The 2-dimensional array representing a grid (1. dimension: row (y-axis), 2. dimension: column (x-axis))
     */
    get grid(): CellWithValue[][];
    /**
     * @param coordinate The coordinate of the cell
     * @returns true if the cell exists in the grid
     */
    cellExists(coordinate: Coordinate): boolean;
    /**
     * @returns All cells in the grid in ascending order by row and column
     */
    get cells(): CellWithValue[];
    /**
     * @param coordinate The coordinate of the cell
     * @returns The cell
     * @throws {CellDoesNotExistInGridError} when the cell does not exist in the grid
     */
    getCell(coordinate: Coordinate): CellWithValue;
    /**
     * @param coordinate The coordinate of the cell
     * @returns The cell or undefined
     */
    findCell(coordinate: Coordinate): CellWithValue | undefined;
    /**
     * @param row The row coordinate
     * @returns The row
     */
    getRow(row: number): Row<Value, CellWithValue>;
    /**
     * @returns All rows of the grid in ascending order
     */
    get rows(): Row<Value, CellWithValue>[];
    /**
     * @param col The column coordinate
     * @returns The column
     */
    getColumn(col: number): Column<Value, CellWithValue>;
    /**
     * @returns All columns of the grid in ascending order
     */
    get columns(): Column<Value, CellWithValue>[];
    /**
     * @param callback A function that should be called, when a cell value of this grid changes
     * @returns a function to unregister the callback
     */
    onCellValueChanged(callback: (event: CellValueChangedEvent<Value>) => void): () => void;
    /**
     * @param gridToAdd The grid that should be added to this grid at the given direction
     * @param addDirection The direction at that the given grid should be placed
     * @returns a new grid
     */
    extend<GridType extends Grid<Value, CellWithValue>>(gridToAdd: Grid<Value, CellWithValue>, addDirection: StraightDirection): GridType;
    /**
     * @param topLeft The coordinate of the top/left where the new grid should begin
     * @param bottomRight The coordinate of the bottom/right where the new grid should end
     * @returns a new grid
     */
    crop<GridType extends Grid<Value, CellWithValue>>(topLeft: Coordinate, bottomRight: Coordinate): GridType;
    /**
     * @param cloneValue A custom function to clone the value of a cell (defaults to copying the value)
     * @returns The cloned grid
     */
    clone<GridType extends Grid<Value, CellWithValue>>(cloneValue?: (value: Value) => Value): GridType;
    protected abstract initializeGrid(options: InitializeGridOptions<Value>): Grid<Value, CellWithValue>;
}
//# sourceMappingURL=Grid.d.ts.map